/* 
 * stunnel - Simple tunnel for any connections using TCP.
 *
 * Copyright (C) 2003-2004  Oskar Andreasson &lt;oanATfrozentuxDOTnet&gt;
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program or from the site that you downloaded it
 * from; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307   USA
 *                                                                          */
  
#include "includes.h"
#include "defines.h"

/* createconnslist
 * Creates the original connslist and adds the first data to it (which will 
 * never be used anyways, so it is simply set to 0).
 *
 * returns 0 on success, or -1 on error. */
void *createconnslist()
{
  struct connslist *root;
  
  root=malloc(sizeof(struct connslist));
  
  root->dstfd = 0;
  root->srcfd = 0;
  root->handled = 0;
  root->next = NULL;
  
  return(root);
}

/* addconnslist
 * Adds a new element to the connslist slist of structures.
 * 
 * returns 0 on success, or -1 on error. */
int addconnslist(struct connslist *root, int srcsockfd, int dstsockfd)
{
  struct connslist *newnode;

#ifdef DEBUG
  printf("Original\naddconnslist.root->srcfd: %i\naddconnslist.root->dstfd: %i\n", root->srcfd, root->dstfd);
#endif

  while(root->next != NULL)
  {
    root = root->next;
  }
  
  newnode = malloc(sizeof(struct connslist));
  newnode->dstfd = dstsockfd;
  newnode->srcfd = srcsockfd;
  newnode->handled = 0;
  newnode->next = NULL;
  root->next = newnode;

#ifdef DEBUG
  printf("New\naddconnslist.root->srcfd: %i\naddconnslist.root->dstfd: %i\n", newnode->srcfd, newnode->dstfd);
#endif

  
  return(0);
}

/* delconnslist
 * Searches for srcsockfd and dstsockfd combo, unless one of them is
 * set to zero, then it looks for only the one with a real value. Finally 
 * deletes what it finds.
 *
 * returns other fd on success, or -1 on error. */
int delconnslist(struct connslist *root, int sock)
{
  struct connslist *prev, *delnode;
  int otherfd;
  while(root->next != NULL)
  {
    prev = root;
    root = root->next;
    
    if(root->srcfd == sock)
    {
      delnode = root;
      prev->next = delnode->next;
      otherfd = delnode->dstfd;
      
      free(delnode);
      return(otherfd);
    }
      
    if(root->dstfd == sock)
    {
      delnode = root;
      prev->next = delnode->next;
      otherfd = delnode->srcfd;
      
      free(delnode);
      return(otherfd);
    }
  }

  return(-1);
}

/* findconnslist
 * Searches for srcsockfd and dstsockfd combo, unless one of them is
 * set to zero, then it looks for only the one with a real value. 
 *
 * returns pointer to connslist struct on success, or NULL on error. */
int findconnslist(struct connslist *root, int sock)
{
  
  while(root->next != NULL)
  {
    root = root->next;
    if(root->srcfd == sock)
      return(root->dstfd);
    if(root->dstfd == sock)
      return(root->srcfd);
  }

  return(-1);
}
