/* 
 * stunnel - Simple tunnel for any connections using TCP.
 *
 * Copyright (C) 2003-2004  Oskar Andreasson &lt;oanATfrozentuxDOTnet&gt;
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program or from the site that you downloaded it
 * from; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307   USA
 *                                                                          */
 
#include "includes.h"
#include "defines.h"

int inc_tcpsock();

int logconn(struct sockaddr_in srchost);
int logdisconn(struct sockaddr_in srchost);

void *createconnslist();
int addconnslist(struct connslist *root, int srcsockfd, int dstsockfd);
int delconnslist(struct connslist *root, int sock);
int findconnslist(struct connslist *root, int sock);

int makeoutconn(struct sockaddr_in dst)
{
  printf("Connecting to %s... ", inet_ntoa(dst.sin_addr));
  int sockfd, yes=1;

  if ((sockfd = inc_tcpsock()) == -1)
  {
    perror("inc_tcpsocket error");
    exit(-1);
  }
      
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) {
    perror("setsockopt");
    exit(1);
  } 

  if (connect(sockfd, (struct sockaddr *)&dst, sizeof(struct sockaddr_in)) == -1)
  {
    perror("listen socket error");
    exit(-1);
  }
  
  printf("Connected.\n");

  return(sockfd);
  
}

int makeinconn(struct sockaddr_in src)
{
  int sockfd, yes=1;

  if ((sockfd = inc_tcpsock()) == -1)
  {
    perror("inc_tcpsocket error");
    exit(-1);
  }
      
  if ((bind(sockfd, (struct sockaddr *)&src, sizeof(struct sockaddr)) == -1))
  {
    perror("bind socket error");
    exit(-1);
  }
  
  if (setsockopt(sockfd,SOL_SOCKET,SO_REUSEADDR,&yes,sizeof(int)) == -1) {
    perror("setsockopt");
    exit(1);
  } 

  if (listen(sockfd, BACKLOG) == -1)
  {
    perror("listen socket error");
    exit(-1);
  }
  
  return(sockfd);

}

int senddata(int newsock, char *msg, size_t len)
{
  
  int sendlen;
  int totsendlen = 0; 
  
  while (len > totsendlen)
  {
    if ((sendlen = send(newsock, msg + totsendlen, len - totsendlen, 0)) == -1)
    {
      perror("sendlen = -1 in senddata");
      return(-1);
    }
    totsendlen = totsendlen + sendlen;
    
#ifdef DEBUG
    printf("Tofd: %i\nLen: %i\nSendlen: %i\nTotsendlen: %i\n", newsock, len, sendlen, totsendlen);
#endif
  }
  
  return(0);
}


/* Main */
int main ()
{
  // Set initial variables.
  int sockfd, tofd;
  struct sockaddr_in srchost, dsthost;
  int straddrsize = sizeof(struct sockaddr_in), newsrcsock=0, newdstsock=0;
  int highfd, i=0, opi=0, sizedata=0;
  fd_set master, readfds;
  char *msg;
  char *data;
  struct connslist *connections;

  // FIXME! temp until I make this into cmdline.
  int dstport = 80;
  char *dsthost_ascii = "192.168.0.9";

  // Initialize variables etc.
  connections = createconnslist();
  
  // fd_sets used through the program.
  FD_ZERO(&master);
  FD_ZERO(&readfds);
  
 // My own listening address
  myaddr.sin_family = AF_INET;
  myaddr.sin_port = htons(LISTEN_PORT);
  myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  memset (&(myaddr.sin_zero), '\0', 8);
  
  // The host to forward traffic to.
  dsthost.sin_family = AF_INET;
  dsthost.sin_port = htons(dstport);
  if(inet_aton(dsthost_ascii, &dsthost.sin_addr) == 0)
    perror("Invalid dsthost_ascii to inet_aton");
  memset (&(dsthost.sin_zero), '\0', 8);
  
  // buffers for incoming/outgoing packets
  data = malloc(sizeof(char)*16384);
  msg = malloc(sizeof(char)*16384);

  // Begin setting up system.
  sockfd = makeinconn(myaddr);


  FD_SET(sockfd, &master);
  
  highfd=sockfd;

  for (;;)
  {
  
    readfds = master;
    if (select(highfd+1, &readfds, NULL, NULL, NULL) == -1)
    {
      perror("Select");
      exit(-1);
    }

    for(i=0;i <= highfd;i++)
    {
      if(FD_ISSET(i, &readfds))
      {
        /* This is a new connection */
        if(i == sockfd)
        {
          logconn(srchost);

          if ((newsrcsock = accept(sockfd, (struct sockaddr *)&srchost, &straddrsize)) == -1)
          {
            perror("accept error");
            exit(-1);
          } else {
            FD_SET(newsrcsock, &master);
            if (newsrcsock > highfd)
              highfd = newsrcsock;
            
            if((newdstsock = makeoutconn(dsthost)) == -1)
              perror("Error makeoutconn(dsthost)");
            FD_SET(newdstsock, &master);
            if (newdstsock > highfd)
              highfd = newdstsock;
              
            if(addconnslist(connections, newsrcsock, newdstsock) == -1)
              perror("addconnslist");

          }
        }
        /* This is an old connection */
        else
        {
          if((sizedata = recv(i, data, sizeof(data), 0)) <= 0)
          {
            // closed by one side.
            if(sizedata == 0)
            {
              if(logdisconn(srchost) == -1)
                perror("logdisconn");
              if((opi = delconnslist(connections, i)) == -1)
                perror("delconnslist");
#ifdef DEBUG
              printf("close(%i)\n", i);
#endif
              if(close(i) == -1)
                perror("close(i)");
#ifdef DEBUG
              printf("close(%i)\n", opi);
#endif
              if(close(opi) == -1)
                perror("close(opi)");

              FD_CLR(i, &master);
              FD_CLR(opi, &master);

            }
            // otherwise an error in receiving.
            else
            {
              printf("sizedata: %i\n", sizedata);
              perror("recv");
              
            }

          }
          // We received data.
          else 
          {
            if((tofd = findconnslist(connections, i)) == -1)
            {
              perror("findconnslist error");
            }
#ifdef DEBUG
            printf("i: %i\ntofd: %i\n", i, tofd);
#endif
            if(senddata(tofd, data, sizedata) != 0)
            {
              perror("senddata error");
              exit(-1);
            }
          }
        
        }
      }
    }
  }
  
  exit(0);
 
}
